#-*-coding:utf-8-*-

"""
Nouvelle définition du profil
=============================

Un profil08 peut comporter des éléments rigides:
  des nombres,
  des tuples,
  des listes de nombres,
  des rectaf, cubicaf, curvaf,

et des éléments élastiques:
  des recta, cubica, curva: '[recta, duree, y0, y1, [puissance]]'
  des ufunc de numpy: '[np.ufunc, duree, y0, y1]'

Les tuples-2 sont chaque fois des limites d'une variation sérielle.

Les fonctions ne doivent donner lieu à un calcul que quand
on connaît la longueur souhaitée.

Le paramètre 'transpose' de 'ProfilBase' n'est pas maintenu.
"""

import numpy as NY
from origine import Origine

import utiles as _u
import profils as _p
import serie

__author__ = "Copyright (c) 2005-07 René Bastian (Wissembourg, FRANCE)"
__date__ = "2008.07.31"
__all__ = ["Profil08", "relatif", "profil08", "recta08"]

def profil08(listeP):
    """
    calcule un vecteur-profil à partir de sa description.
    'P' : liste des composants du profil
    06.12.07 ; env. 12 % plus rapide que la précédente version.
    La version 'liste de listes' a l'avantage de pouvoir être
    utilisée par 'ProfilBase'.
    """
    pR = []
    for e in listeP:
        pR.append(e[0](e[1:]))        
    return NY.concatenate(pR)

def recta08(D):
    """
    trace une droite allant de D[1] à D[2] en une durée D[0]
    adaptation à numpy de 'racta'
    """
    try:
        duree, y0, y1, puiss = D
    except StandardError:
        duree, y0, y1 = D
        puiss = 1.0
    nd = int(duree * _u.sr)
    if nd <= 1:
        return NY.array([(y0 + y1)/2])
    if y0 == y1:
        return NY.ones(nd) * y0
    if puiss != 1.0:
        u = NY.arange(0.0, 1.0, 1.0/nd)
        u **= puiss
        u *= y1 - y0
        return y0 + u
    else:
        return y0 + NY.arange(0.0, 1.0, 1.0/nd)*(y1 - y0)

def relatif(duree, aprofil, f):
    """
    'duree' : durée relative
    'aprofil' : profil-liste 
    'f' : fonction 
    exemple:
       relatif(2.0, [_p.cubica, None, y0, p0, y1, p1],  NY.sin)
    """
    v = aprofil[0](duree, aprofil[2:])
    return f(v)

class Profil08(Origine):
    """
Nouvelle définition du profil
=============================

Un profil08 peut comporter des éléments rigides:
  des nombres,
  des tuples,
  des listes de nombres,
  des rectaf, cubicaf, curvaf,

et des éléments élastiques:
  des recta, cubica, curva: '[recta, duree, y0, y1, [puissance]]'
  des ufunc de numpy: '[np.ufunc, duree, y0, y1]'

Les tuples-2 sont chaque fois des limites d'une variation sérielle.

Les fonctions ne doivent donner lieu à un calcul que quand
on connaît la longueur souhaitée.

Le paramètre 'transpose' de 'ProfilBase' n'est pas maintenu.
"""
    def __init__(self, listeP, couplealea=None):
        "procède à l'instanciation"
        Origine.__init__(self)
        self.p = None
        self.duree  = None
        self.incr = 1./_u.sr
        self.listeP = listeP
        if couplealea:
            self.alea = serie.Serie002(*couplealea)
        else:
            self.alea = serie.Serie002()
        self.nrigide = 0
        self.telastique = 0.0
        for px in listeP:
            if isinstance(px, float) or isinstance(px, int) \
                    or isinstance(px, long):
                self.nrigide += 1
            elif isinstance(px, tuple):
                self.nrigide += len(tuple)
            elif isinstance(px, list):
                if (px[0] in [_p.rectaf, _p.curvaf, _p.cubicaf]):
                    nelement = int(px[1] * _u.sr)
                    self.nrigide += nelement
                elif (px[0] in [_p.recta, _p.curva, _p.cubica]):
                    self.telastique += px[1]
                elif px[0] == relatif:
                    self.telastique += px[1]
                else:                    
                    self.nrigide += len(px)
            elif isinstance(px, NY.ndarray):
                self.nrigide += len(px)
            else:
                self.telastique += px[1]
        self.trigide = float(self.nrigide) / _u.sr

    def __call__(self):
        pass
    
    def temps(self, duree, alea=None):
        """calcule un vecteur de durée 'duree'.

        Si les durées élastiques sont à variation sérielles, 'duree'
        n'a qu'une valeur approximative.

        Les cas 5 et 6 sont possibles mais sémantiquement pas stabiliés.
        Il est plus logique de fabriquer le vecteur à l'extérieur
        p. ex. dans une sous-classe) et de l'intégrer (cas 4)
        """        
        if duree < self.trigide:
            s = "Profil08:temps:duree %f << trigide %f" % (duree, self.trigide)
            print s
            return None
        if not alea:
            alea = self.alea
        tvar = duree - self.trigide
        try:
            rapport = tvar / self.telastique
        except ZeroDivisionError:
            rapport = 1.0
        R = []
        for px in self.listeP:
            if isinstance(px, float) or isinstance(px, int) \
                    or isinstance(px, long):
                R.append([px])
            elif isinstance(px, tuple):
                R.append(list(px))
            elif isinstance(px, list): 
                if (px[0] in [_p.rectaf, _p.curvaf, _p.cubicaf]):
                    R.append(px[0](px[1:]))
                elif (px[0] in [_p.recta, _p.curva, _p.cubica]):
                    xduree = px[1] * rapport
                    pp = px[0]([xduree] + px[2:])
                    R.append(pp)
                elif px[0] == relatif:
                    xduree = px[1] * rapport
                    pp = relatif(xduree, *px[2:])
                    R.append(pp)
                else:
                    R.append(px)
            elif isinstance(px, NY.ndarray): 
                R.append(px)
            elif isinstance(px[0], NY.ufunc):
                
                print "experimental", px
                xduree = px[1] * rapport
                nx = int(xduree * _u.sr)
                dv = (px[3] - px[2]) / nx
                v = NY.arange(px[2], px[3], dv)
                R.append(px[0](v))
            else: 
                
                print "experimental", px
                xduree = px[1] * rapport 
                pp = px[0](*px[2:])
                R.append(pp)
        R = NY.concatenate(R)
        self.p = R
        self.duree = float(len(R)) / _u.sr
        return R

    def samples(self, n, alea=None):
        """calcule un vecteur de 'n' samples.
        Il faut que ce 'n' soit absolument respecté.
        NE FONCTIONNE PAS
        """
        if n < self.nrigide:
            print "Profil08:samples: 'n' trop faible", n, self.nrigide
            return None
        else:
            duree = float(n) / _u.sr
            w = self.temps(duree, alea)
            if w is None:
                print duree, n
            
            while len(w) > n:
                duree -= self.incr
                w = self.temps(duree, alea)
            while len(w) < n:
                duree += self.incr
                w = self.temps(duree, alea)
            return w[:n]

    def extraitsamples(self, w, n0, dn):
        """prend le fragment w[n0 : n0+dn] et lui impose le profil."""
        wx = w[n0:n0+dn]
        e = self.samples(dn)
        return wx * e

    def extrait(self, w, n0=None, dn=None, t=None, dt=None, brut=False):
        """prend 'w' ou un fragment 'w[n0 : n0+dn]' et lui impose le profil."""
        if n0 and dn:
            pass
        elif n0 and dt:
            dn = int(dt * _u.sr)
        elif t and dt:
            n0 = int(t * _u.sr)
            dn = int(dt * _u.sr)
        elif t and dn:
            n0 = int(t * _u.sr)
        else:
            n0 = 0
            dn = len(w)
        try:
            wx = w[n0:n0+dn]
        except StandardError:
            print "Profil08.extrait flanche"
        
        if brut:
            return wx
        else:
            e = Profil08(self.listeP).samples(dn)
            
            return wx * e

    def inverse(self, w, n0=None, dn=None, t=None, dt=None, invbrut=True):
        "retourne la forme rétrograde"
        wx = self.extrait(w, n0=n0, dn=dn, t=t, dt=dt, brut=True)
        if invbrut:
            e = Profil08(self.listeP).samples(dn)
            return wx[::-1] * e
        else:
            return wx[::-1]

    def symetrique(self, w, n0=None, dn=None, t=None, dt=None, symbrut=False):
        "retourne l'inversion"
        wx = self.extrait(w, n0=n0, dn=dn, t=t, dt=dt, brut=True)
        wx = NY.concatenate((wx, wx[::-1]))
        if symbrut:
            return wx
        else: 
            e = Profil08(self.listeP).samples(len(wx))            
            return wx * e


